#include "lib_indicator.h"
#include "lib_dmx_in.h"

#define  F_CPU	(8000000UL)

#include <stdlib.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>


		 uint8_t   Settings;
		 uint8_t   Drive;				//type of driver (0= bipolar; 1= unipolar; else= step/dir)
volatile uint8_t   Flags;
		 uint8_t   IndCnt= 10;
		 uint8_t   SysCnt= 4;
		 uint8_t   PwmCnt= 0;			//PWM bit mask counter
		 uint8_t   PwmSort[4];			//buffer for sorted brightness levels (ascending)
		 uint8_t   PwmBuf[4];			//buffer for raw pwm data
		 uint8_t   StrobeCnt;


		struct axis{				//all private data for X and Y axes
		   uint16_t   Ist;				//current position
		   uint16_t	  Soll;				//target position
		   uint16_t   Max;				//max position
		   uint8_t    Accel;			//acceleration (hi val= slow)
		   int8_t	  Speed;      		//current speed (raw)
		   uint16_t   StpRate;
  volatile uint8_t	  Tick;				//Counter for speed calc in main
		   };
		   
		struct T_PWM{				//final data for PWM ISR
			uint8_t   Time[5];			//quench time
			uint8_t	  Quench[5];		//quench pattern for PortC
			uint8_t	  Fire;
			};
		   
		struct   axis       X;
		struct   axis       Y;
		struct	 T_PWM		PwmA;		//PWM isr data with backbuffer
		struct	 T_PWM		PwmB;
		struct	 T_PWM		*PwmRead=  &PwmA;
		struct	 T_PWM		*PwmWrite= &PwmB;


enum {TICK, CFG_WRT_EE, CFG_X0, CFG_Y0, PWM_READY, PWM_ON};					//Flags

#define OFFSET			(12)
#define IND_TIME		(40)
#define SYS_TIME		(3)
#define FLASH_DURATION	(5)				//while strobing

#define SET_DRV0	    (0)				//Settings (bits)
#define SET_DRV1	    (1)
#define SET_HOME0	    (2)
#define SET_HOME1	    (3)
#define SET_LED_INV	    (6)
#define SET_HIRES	    (7)

#define EE_CHECK		(0)
#define EE_X_MAX_H		(1)				//EEPROM usage
#define EE_X_MAX_L		(2)
#define EE_Y_MAX_H		(3)
#define EE_Y_MAX_L		(4)
#define EE_X_ACC		(5)
#define EE_Y_ACC		(6)
#define EE_SETTINGS		(7)


const uint8_t biStep[] PROGMEM =	{								//bipolar half-steps (DCBA DCBA)
									0b00010001,
									0b01010101,
									0b01000100,
									0b01100110,
									0b00100010,
									0b10101010,
									0b10001000,
									0b10011001,
									0b00010001
								};


const uint8_t uniStep[] PROGMEM =	{								//unipolar half-steps
									0b00010001,
									0b00110011,
									0b00100010,
									0b01100110,
									0b01000100,
									0b11001100,
									0b10001000,
									0b10011001,
									0b00010001
								};



const uint16_t Ramp[] PROGMEM = {								//acceleration, deceleration (step durations)
27145,  20829,  17560,  15471,  13987,  12862,  11972,  11244,  10635,  10115,  9665,  9270,  8920,  8606,
8324,  8068,  7834,  7619,  7421,  7238,  7067,  6908,  6760,  6620,  6489,  6366,  6249,  6138,  6033,  5933,
5839,  5748,  5662,  5579,  5500,  5424,  5351,  5281,  5214,  5149,  5087,  5026,  4968,  4912,  4858,  4805,
4754,  4705,  4657,  4611,  4566,  4522,  4480,  4439,  4398,  4359,  4321,  4284,  4248,  4213,  4178,  4145,
4112,  4080,  4049,  4018,  3988,  3959,  3931,  3903,  3875,  3848,  3822,  3796,  3771,  3746,  3722,  3698,
3675,  3652,  3630,  3608,  3586,  3565,  3544,  3523,  3503,  3483,  3464,  3444,  3426,  3407,  3389,  3371,
3353,  3336,  3319,  3302,  3285,  3269,  3252,  3237,  3221,  3205,  3190,  3175,  3160,  3146,  3131,  3117,
3103,  3089,  3076,  3062,  3049,  3036,  3023,  3010,  2998,  2985,  2973,  2961,  2949,  2937,  2925,	
};

const uint8_t StrobeRate[] PROGMEM = {
	151, 148, 146, 144, 141, 139, 137, 134, 132, 130, 127, 125, 123, 121, 118, 116,
	114, 112, 110, 108, 106, 104, 102,  99,  97,  95,  93,  92,  90,  88,  86,  84,
	82,  80,  78,  76,  75,  73, 71,   69,  68,  66,  64,  63,  61,  59,  58,  56, 55, 53,
	51,  50,  48,  47,  46,  44, 43,   41,  40,  39,  37,  36,  35,  33,  32,  31, 30, 29,
	27,  26,  25,  24,  23,  22, 21,   20,  19,  18,  17,  16,  15,  15,  14,  13, 12, 12,
	11,  11,  10,  10,   9,   9,  8,    8,   7,   7,   6,   6,   5,   5,   5,   4,  4,  4
};


// *************** Stepper Homing ****************
uint8_t home_x(void)
{
uint8_t tmp= 0;
if (PIND &(1<<PD4))
   {
   if (Drive == 0)											// bipolar driver
	   {
	   tmp  = pgm_read_byte(&biStep[(--X.Ist &7)]);
	   tmp &= 0b00001111;
	   }
	else if (Drive == 1)									// unipolar driver
	   {
	   tmp  = pgm_read_byte(&uniStep[(--X.Ist &7)]);
	   tmp &= 0b00001111;
	   }
	else
	   {													// step / dir interface
	   PORTA |=  (1<<PA0);									//set step pin (cleared in home_stepper by PORTA= tmp...)
	   --X.Ist;
	   _delay_us(10);
	   tmp= 0;
	   }
   }
else
   {
   X.Ist &= 7;												//end stop detected (closed loop reference)
   X.Ist += (OFFSET -7);
   }
return (tmp);	
}

uint8_t home_y(void)
{
uint8_t tmp= PORTA;
if (PIND &(1<<PD5))
   {
   if (Drive == 0)											// bipolar driver
	   {
	   tmp  = pgm_read_byte(&biStep[(--Y.Ist &7)]);
	   tmp &= 0b11110000;
	   }
	else if (Drive == 1)									// unipolar driver
	   {
	   tmp  = pgm_read_byte(&uniStep[(--Y.Ist &7)]);
	   tmp &= 0b11110000;
	   }
	else
	   {													// step / dir interface
	   PORTA |=  (1<<PA2);									//set step pin (cleared in home_stepper by PORTA= tmp...)
	   --Y.Ist;
	   _delay_us(10);
	   tmp= 0;
	   }
   }
else
   {
	Y.Ist &= 7;												//end stop detected (closed loop reference)
	Y.Ist += (OFFSET -7);
   }
return (tmp);
}

void home_stepper(uint8_t mode)
{
//position reset (homing)
uint8_t tmpA= 0, tmpB= 0;
X.Ist= X.Max +20 +OFFSET;
Y.Ist= Y.Max +20 +OFFSET;
uint16_t delay;
if (mode == 0)															//simultaneous homing
   {
   while ((X.Ist != OFFSET) || (Y.Ist != OFFSET))
      {
	   if (X.Ist > OFFSET) tmpA= home_x();
	   if (Y.Ist > OFFSET) tmpB= home_y();
	   tmpA |= tmpB;													//merge axes
	   PORTA= tmpA;
	   wdt_reset();
	   if (Drive < 2) delay= (X.Accel<<1);
	   else           delay= (X.Accel>>2);								//8x speed in microstepping mode
	   while (delay != 0)
		   {
		   _delay_us(35);
		   delay--;
		   } 
	   }
	}
else if (mode == 1)
    {																	//X first
	while (X.Ist > OFFSET)
	    {
		tmpA= home_x();
		PORTA= tmpA;
		wdt_reset();
		if (Drive < 2) delay= (X.Accel<<1);
		else           delay= (X.Accel>>2);								//8x speed in microstepping mode
		while (delay != 0)
		   {
			_delay_us(35);
			delay--;
		   }
	    }
	while (Y.Ist > OFFSET)
	    {
		tmpB= home_y();
	    PORTA= tmpB;
	    wdt_reset();
	    if (Drive < 2) delay= (X.Accel<<1);
	    else           delay= (X.Accel>>2);								//8x speed in microstepping mode
	    while (delay != 0)
	       {
		   _delay_us(35);
		   delay--;
	       }
	    }
	}
else 
    {
	while (Y.Ist > OFFSET)												//Y first
	    {
		tmpB= home_y();
		PORTA= tmpB;
		wdt_reset();
		if (Drive < 2) delay= (X.Accel<<1);
		else           delay= (X.Accel>>2);								//8x speed in microstepping mode
		while (delay != 0)
		    {
			_delay_us(35);
			delay--;
		    }
	    }
	while (X.Ist > OFFSET)
	    {
		tmpA= home_x();
		PORTA= tmpA;
		wdt_reset();
		if (Drive < 2) delay= (X.Accel<<1);
		else           delay= (X.Accel>>2);								//8x speed in microstepping mode
		while (delay != 0)
		    {
			_delay_us(35);
			delay--;
		    }
	    }
    }
}


// *************** Initialization ****************
void init_system(void)
{
uint8_t tmp;
wdt_reset();
wdt_enable(WDTO_500MS);											//enable watchdog (Timeout= 0.5s)
wdt_reset();

DDRA=  0xFF;													//low outputs
PORTA= 0;

DDRD=  0b10000110;												//DMX, spare, red LED1 on
PORTD= 0b01111010;

if (eeprom_read_byte(EE_CHECK) == 1)
	{
	X.Max= (uint16_t)(eeprom_read_byte(EE_X_MAX_H)<<8) +(uint16_t)eeprom_read_byte(EE_X_MAX_L);
	Y.Max= (uint16_t)(eeprom_read_byte(EE_Y_MAX_H)<<8) +(uint16_t)eeprom_read_byte(EE_Y_MAX_L);
	X.Accel= eeprom_read_byte(EE_X_ACC);
	Y.Accel= eeprom_read_byte(EE_Y_ACC);
	Settings= eeprom_read_byte(EE_SETTINGS);
	}
else
	{
	X.Max= 400;
	Y.Max= 400;
	X.Accel= 100;
	Y.Accel= 100;
	Settings= (1<<SET_HIRES);					//use 16bit resolution as default, bipolar driver
	eeprom_write_byte(EE_CHECK, 1);
	eeprom_write_byte(EE_X_MAX_H, (X.Max>>8)&0xFF);
	eeprom_write_byte(EE_X_MAX_L, (X.Max)&0xFF);
	eeprom_write_byte(EE_Y_MAX_H, (Y.Max>>8)&0xFF);
	eeprom_write_byte(EE_Y_MAX_L, (Y.Max)&0xFF);
	eeprom_write_byte(EE_X_ACC, X.Accel);
	eeprom_write_byte(EE_Y_ACC, Y.Accel);
	eeprom_write_byte(EE_SETTINGS, Settings);
	}

tmp= 0;											//get driver configuration
if (Settings &(1<<SET_DRV0)) tmp  = 1;
if (Settings &(1<<SET_DRV1)) tmp += 2;
Drive= tmp;

tmp= 0;											//get home sequence
if (Settings &(1<<SET_HOME0)) tmp  = 1;
if (Settings &(1<<SET_HOME1)) tmp += 2;
home_stepper(tmp);

TCCR0  = (1<<CS01)|(1<<CS00);									//set T0 @clk/64 (8bit fast PWM with 488Hz)
TCCR1A = 0;
if (Drive < 2) TCCR1B = (1<<CS11);								//set T1 @clock/8 for half step operation
else		   TCCR1B = (1<<CS10);								//set T1 @clock for microstepping
TIMSK  = (1<<OCIE1A)|(1<<OCIE1B)|(1<<OCIE0);					//enable compare interrupts (stepper control, PWM, sys clock)

PwmA.Time[4]= 0xFF;								//save exit
PwmB.Time[4]= 0xFF;
PwmA.Fire= 0;									//all ch off
PwmB.Fire= 0;
uint8_t i;
for (i=0; i< 5; i++)
   {
	PwmA.Quench[i]= 0;
	PwmA.Quench[i]= 0;
   }
for (i=0; i< 4; i++)
   {
	PwmBuf[i]= 0;
   }
Flags = (1<<PWM_ON);

init_ind();
init_DMX_RX();
get_dips();
}


// *************** DMX Configuration ****************
//ch1 = val1 | val1_H
//ch2 = val2 | val1_L
//ch3 = val3 | val2_H
//ch4 = val4 | val2_L
//ch5 = ee write cmd
//ch6 = parameter to configure

void config_basic(void)
{
uint8_t tmp;
if (DmxRxField[4] < 128)
   {
   Flags &= ~(1<<CFG_WRT_EE); 
   return;
   }	   
if (!(Flags &(1<<CFG_WRT_EE)))
   {
   Flags |= (1<<CFG_WRT_EE);
   tmp= 0;								//buffer for settings
   //** Drive **
   Drive= 0;							
   if (DmxRxField[0] &(1<<6))
      {
	  Drive = 1;
	  tmp  |= (1<<SET_DRV0);  
	  }
   if (DmxRxField[0] &(1<<7))
      {
	  Drive += 2;
	  tmp   |= (1<<SET_DRV1);
      }
   // ** homing sequence **
   if (DmxRxField[1] &(1<<6)) tmp |= (1<<SET_HOME0);
   if (DmxRxField[1] &(1<<7)) tmp |= (1<<SET_HOME1);
   // ** resolution **
   if (DmxRxField[2] &(1<<7)) tmp |= (1<<SET_HIRES);
   Settings= tmp;
   eeprom_write_byte (EE_SETTINGS, tmp);											
   eeprom_write_byte (EE_CHECK, 1);											//mark EE data available
   IndFlags |= (1<<DATA_REFRESHED); 
   }  	
}

void config_range(void)
{
X.Soll  = ((uint16_t)DmxRxField[0]<<8) +(uint16_t)DmxRxField[1];			//get 16bit positions from DMX
Y.Soll  = ((uint16_t)DmxRxField[2]<<8) +(uint16_t)DmxRxField[3];
if (X.Soll < (0xFFFF -OFFSET)) X.Soll += OFFSET;
if (Y.Soll < (0xFFFF -OFFSET)) Y.Soll += OFFSET;
X.StpRate= pgm_read_word(&Ramp[0]);											//use min speed; no acceleration
if (X.Soll == OFFSET)
   {
	X.Speed= -1;															//use 0 for homing
	Flags |= (1<<CFG_X0);  
   }
else
   {
	if (Flags &(1<<CFG_X0))
	   {
		Flags &= ~(1<<CFG_X0);												//reset position, if we come from homing
		X.Ist  = OFFSET;
	   }
	if      (X.Soll > X.Ist) X.Speed=  1;									//move slowly to target position
	else if (X.Soll < X.Ist) X.Speed= -1;
	else                     X.Speed=  0;
   }
Y.StpRate= X.StpRate;
if (Y.Soll == OFFSET)
   {
	Y.Speed= -1;															
	Flags |= (1<<CFG_Y0);
   }
else
   {
	if (Flags &(1<<CFG_Y0))
	   {
		Flags &= ~(1<<CFG_Y0);												
		Y.Ist  = OFFSET;
	   }
	if      (Y.Soll > Y.Ist) Y.Speed=  1;
	else if (Y.Soll < Y.Ist) Y.Speed= -1;
	else                     Y.Speed=  0;
   }
if (DmxRxField[4] < 128)
   {
   Flags &= ~(1<<CFG_WRT_EE); 
   return;
   }	   
if (!(Flags &(1<<CFG_WRT_EE)))
   {
   Flags |= (1<<CFG_WRT_EE);
   X.Max= X.Soll;
   Y.Max= Y.Soll;
   eeprom_write_byte (EE_X_MAX_H, DmxRxField[0]);								//store max positions
   eeprom_write_byte (EE_X_MAX_L, DmxRxField[1]);
   eeprom_write_byte (EE_Y_MAX_H, DmxRxField[2]);
   eeprom_write_byte (EE_Y_MAX_L, DmxRxField[3]);												
   eeprom_write_byte (EE_CHECK, 1);												//mark EE data available
   IndFlags |= (1<<DATA_REFRESHED);   
   }  	
}


void config_acc(void)
{
//** get 8bit target positions **
uint32_t tmp;
tmp= (uint32_t)DmxRxField[2];										
tmp= OFFSET +((tmp * X.Max)>>8);
X.Soll= tmp;
tmp= (uint32_t)DmxRxField[3];													
tmp= OFFSET +((tmp * Y.Max)>>8);
Y.Soll= tmp;
//** get accelerations **
uint8_t acc;														
acc= (0xFF -DmxRxField[0]);
if (acc < 50) acc=50;
X.Accel= acc;
acc= (0xFF -DmxRxField[1]);
if (acc < 50) acc=50;
Y.Accel= acc;

if (DmxRxField[4] < 128)
	{
	Flags &= ~(1<<CFG_WRT_EE);
	return;
	}
if (!(Flags &(1<<CFG_WRT_EE)))
	{
	Flags |= (1<<CFG_WRT_EE);
	eeprom_write_byte (EE_X_ACC, X.Accel);						//store max positions
	eeprom_write_byte (EE_Y_ACC, Y.Accel);
	eeprom_write_byte (EE_CHECK, 1);							//mark EE data available
	IndFlags |= (1<<DATA_REFRESHED);
	}
}

// *************** Stepper speed control ****************
void check_speed(struct axis* Stp)
{
if (abs(Stp->Soll -Stp->Ist) <= abs(Stp->Speed))				//decel speed short before target position
	{
	if      (Stp->Speed < 0) Stp->Speed++;
	else if (Stp->Speed > 0) Stp->Speed--; 				
	}

else if (Stp->Ist < Stp->Soll)									//shall move in positive direction
	{
	if (Stp->Speed < 120) Stp->Speed++; 						//accel speed
	}

else if (Stp->Ist > Stp->Soll)									//shall move in negative direction
	{
	if (Stp->Speed > -120) Stp->Speed--; 						//accel speed
	}

uint32_t tmp= (uint32_t)pgm_read_word(&Ramp[abs(Stp->Speed)]);
Stp->StpRate= ((tmp * Stp->Accel)>>8);
}



// *************** Target Positions ****************
void eval_X_TargetPos(void)
{
uint32_t tmp;
if (Settings &(1<<SET_HIRES))
   {																//16bit mode
   tmp= ((uint32_t)DmxRxField[0]<<8) +(uint32_t)DmxRxField[1];
   tmp= OFFSET +((tmp * X.Max)>>16);
   if (tmp != X.Soll)
	   {
	   X.Soll= tmp;
	   IndFlags |= (1<<DATA_REFRESHED);
	   }
   }
else
   {
	tmp= (uint32_t)DmxRxField[0];
	tmp= OFFSET +((tmp * X.Max)>>8);
	if (tmp != X.Soll)
	    {
		X.Soll= tmp;
		IndFlags |= (1<<DATA_REFRESHED);
	    } 
   }
}


void eval_Y_TargetPos(void)
{
uint32_t tmp;
if (Settings &(1<<SET_HIRES))
	{																//16bit mode
	tmp= ((uint32_t)DmxRxField[2]<<8) +(uint32_t)DmxRxField[3];
	tmp= OFFSET +((tmp * Y.Max)>>16);
	if (tmp != Y.Soll)
		{
		Y.Soll= tmp;
		IndFlags |= (1<<DATA_REFRESHED);
		}
	}
else
	{
	tmp= (uint32_t)DmxRxField[1];
	tmp= OFFSET +((tmp * Y.Max)>>16);
	if (tmp != Y.Soll)
		{
		Y.Soll= tmp;
		IndFlags |= (1<<DATA_REFRESHED);
		}
	}
}


// *************** PWM Control (for LED outputs) ****************
static void sort(void)													//ascending bubble sort (simple & short but slow...)
{
uint8_t a, b, i, k;
if (Settings &(1<<SET_HIRES)) i= 5;
else                          i= 3;
for (k=0; k<4; k++)
    {
	a= DmxRxField[i];
	if (a != PwmBuf[k])
	   {
		PwmBuf[k]= a;
		IndFlags |= (1<<DATA_REFRESHED);	//mark change of pwm data (flash green)
	   }
	i++;
	}	
for (k=0; k<4; k++)
    {
	a= PwmBuf[k];
	if (a <= 1) a= 0xFF;		//if ch is off, there is nothing to quench
	PwmSort[k]= a;
	}
for (i= 3; i > 0; i--)			//sort quench times ascending
	{
	for (k= 0; k < i; k++)
		{
		a= PwmSort[k];
		b= PwmSort[k +1];
		if (a > b)
			{
			PwmSort[k]= b;
			PwmSort[k +1]= a;
			}
		}
	}
}

static void build_fire_mask(void)										//fire all ch > 1 on roll-over (T0 comp @0xFF)
{
uint8_t buf= 0;
if (PwmBuf[0] > 1) buf  =  (1<<PA4);
if (PwmBuf[1] > 1) buf |=  (1<<PA5);
if (PwmBuf[2] > 1) buf |=  (1<<PA6);
if (PwmBuf[3] > 1) buf |=  (1<<PA7);
PwmWrite->Fire= buf;
}

static void build_quench_mask(void)										//quench ch on their pwm time (T0 comp)
{
uint8_t ch, comp, buf;
for (ch= 0; ch < 4; ch++)
	{
	comp= PwmSort[ch];
	if (comp < 0xFE)
		{
		buf= 0xFF;
		if (PwmBuf[0] <= comp) buf &= ~(1<<PA4);
		if (PwmBuf[1] <= comp) buf &= ~(1<<PA5);
		if (PwmBuf[2] <= comp) buf &= ~(1<<PA6);
		if (PwmBuf[3] <= comp) buf &= ~(1<<PA7);	
		PwmWrite->Quench[ch]= buf;
		PwmWrite->Time[ch]= comp;										//eval quench time for current pattern
		}
	else
		{
		PwmWrite->Quench[ch]= 0xFF;
		PwmWrite->Time[ch]= 0xFF;
		}

	}
}


void kill_pwm(void)
{
uint8_t ch;
for (ch=0; ch<4; ch++)
	{
	PwmWrite->Quench[ch]= 0b00001111;
	PwmWrite->Time[ch]  = 0xFF;
	}
PwmWrite->Fire= 0;
Flags |= (1<<PWM_READY);												//refresh finished 
}

void refresh_pwm(void)
{
sort();
build_fire_mask();
build_quench_mask();
Flags |= (1<<PWM_READY);												//refresh finished   
}


void eval_strobe(uint8_t rate)
{
if (rate >= 50)
	{
	if (--StrobeCnt == 0)										//strobe enabled
		{
		if (Flags &(1<<PWM_ON))
			{
			rate -= 50;
			rate  = (rate>>1);									//(table has only 105 values)
			StrobeCnt= pgm_read_byte(&StrobeRate[rate]);
			Flags &= ~(1<<PWM_ON);								//blank
			}
		else
			{
			StrobeCnt= FLASH_DURATION;
			Flags |= (1<<PWM_ON);								//flash
			}
		}
	}
	else
	{
	StrobeCnt= FLASH_DURATION;
	Flags |= (1<<PWM_ON);										//enable PWM
	}
}

int main(void)
{
cli();
init_system();
sei();
for(;;)
	{
	if (PIND &(1<<PD6))
	   {									//Spare3 open = normal operation
		if (X.Tick != 0)
		   {
		   X.Tick -= 1; 
		   eval_X_TargetPos();
		   check_speed(&X);  
	       }
		if (Y.Tick != 0)
	       {
		   Y.Tick -= 1;
		   eval_Y_TargetPos();
		   check_speed(&Y);
		   }
	   }
	else
	   {									//Spare3 set = use DMX for configuration
		if (DmxRxField[5] < 85)
		   {								//first step: basic settings
		   config_basic();
		   }
		else if (DmxRxField[5] < 170)
		   {								//second step: max positions
		   config_range();
		   }
		else
		   {								//third step: accelerations
		   config_acc();								
		   if (X.Tick != 0)
		      {
			   X.Tick -= 1;
			   check_speed(&X);
		      }
		   if (Y.Tick != 0)
		      {
			   Y.Tick -= 1;
			   check_speed(&Y);
		      }
		   }
	   }
	if (Flags &(1<<TICK))
		{									//called with 488Hz (by PWM ISR)
		Flags &= ~(1<<TICK);
		if (SysCnt-- == 0)
		   {
			SysCnt= SYS_TIME;				//scale down to ~200Hz
			wdt_reset();
			if (Drive == 3)
			   {
			   if (Settings &(1<<SET_HIRES)) eval_strobe(DmxRxField[4]);
			   else                          eval_strobe(DmxRxField[2]);
			   if (Flags &(1<<PWM_ON)) refresh_pwm();
			   else					   kill_pwm();
			   }
			else 
			   {
			   kill_pwm();
			   }
		   }
		if (IndCnt-- == 0)
			{
			IndCnt= IND_TIME;				//scale down to ~10Hz
			get_dips();
			indicate();
			}
		}
	};
}



// *************** Stepper ISRs ****************
ISR (TIMER1_COMPA_vect)									//X-axis
{
X.Tick++;
uint8_t tmp;
uint8_t buf;
if (X.Speed > 0)											//move in pos direction
   {
   if (Drive == 0)											// bipolar driver
	  {
	  tmp  = pgm_read_byte(&biStep[(++X.Ist &7)]);
	  buf  = PORTA;
	  buf &= 0b11110000;
	  tmp &= 0b00001111;
	  tmp |= buf;
	  PORTA= tmp;
	  }
	else if (Drive == 1)									// unipolar driver
	  {
	  tmp= pgm_read_byte(&uniStep[(++X.Ist &7)]);
	  buf  = PORTA;
	  buf &= 0b11110000;
	  tmp &= 0b00001111;
	  tmp |= buf;
	  PORTA= tmp;
	  }
	else
	  {														// step / dir interface
	  PORTA |= (1<<PA1);
	  PORTA |= (1<<PA0);
	  ++X.Ist;
	  _delay_us(10);
	  PORTA &= ~(1<<PA0);
	  }
   }
else if (X.Speed < 0)										//move in neg direction
   {
	if (Drive == 0)											// bipolar driver
	   {
		tmp  = pgm_read_byte(&biStep[(--X.Ist &7)]);
		buf  = PORTA;
		buf &= 0b11110000;
		tmp &= 0b00001111;
		tmp |= buf;
		PORTA= tmp;
	   }
	else if (Drive == 1)									// unipolar driver
	   {
		tmp= pgm_read_byte(&uniStep[(--X.Ist &7)]);
		buf  = PORTA;
		buf &= 0b11110000;
		tmp &= 0b00001111;
		tmp |= buf;
		PORTA= tmp;
	   }
	else
	   {														// step / dir interface
		PORTA &= ~(1<<PA1);
		PORTA |=  (1<<PA0);
		--X.Ist;
		_delay_us(10);
		PORTA &= ~(1<<PA0);
	   }
   }
OCR1A += X.StpRate;
}


ISR (TIMER1_COMPB_vect)									//Y-Axis
{
Y.Tick++;
uint8_t tmp;
uint8_t buf;
if (Y.Speed > 0)											//move in pos direction
   {
	if (Drive == 0)											// bipolar driver
	   {
		tmp  = pgm_read_byte(&biStep[(++Y.Ist &7)]);
		buf  = PORTA;
		buf &= 0b00001111;
		tmp &= 0b11110000;
		tmp |= buf;
		PORTA= tmp;
	   }
	else if (Drive == 1)									// unipolar driver
	   {
		tmp= pgm_read_byte(&uniStep[(++Y.Ist &7)]);
		buf  = PORTA;
		buf &= 0b00001111;
		tmp &= 0b11110000;
		tmp |= buf;
		PORTA= tmp;
	   }
	else
	   {														// step / dir interface
		PORTA |= (1<<PA3);
		PORTA |= (1<<PA2);
		++Y.Ist;
		_delay_us(10);
		PORTA &= ~(1<<PA2);
	   }
   }
else if (Y.Speed < 0)										//move in neg direction
   {
	if (Drive == 0)											// bipolar driver
	   {
		tmp  = pgm_read_byte(&biStep[(--Y.Ist &7)]);
		buf  = PORTA;
		buf &= 0b00001111;
		tmp &= 0b11110000;
		tmp |= buf;
		PORTA= tmp;
	   }
	else if (Drive == 1)									// unipolar driver
	   {
		tmp= pgm_read_byte(&uniStep[(--Y.Ist &7)]);
		buf  = PORTA;
		buf &= 0b00001111;
		tmp &= 0b11110000;
		tmp |= buf;
		PORTA= tmp;
	   }
	else
	   {														// step / dir interface
		PORTA &= ~(1<<PA3);
		PORTA |=  (1<<PA2);
		--Y.Ist;
		_delay_us(10);
		PORTA &= ~(1<<PA2);
	   }
   }
OCR1B += Y.StpRate;										//set next step time
}

// *************** PWM & SysTick ISR ****************
ISR (TIMER0_COMP_vect)
{
uint8_t buf;
uint8_t nxt;  
if (Drive != 3)
   {
	Flags |= (1<<TICK);
	OCR0= 0xFF;
	return;   
   } 
   
nxt= OCR0;    
if (nxt == 0xFF)
   {
   buf= Flags;
   if (buf &(1<<PWM_READY))											//change backbuffer
	  {
	  buf &= ~(1<<PWM_READY);
	  struct	 T_PWM	*write_buf= PwmWrite;
	  PwmWrite= PwmRead;
	  PwmRead= write_buf;
	  }
   buf |= (1<<TICK);
   Flags= buf;

   buf   = PORTA &0b00001111;
   buf  |= PwmRead->Fire;
   PORTA = buf;

   PwmCnt= 0;
   nxt  = PwmRead->Time[0];  
   }
else
   {
   do 
       {
	   buf    = PwmRead->Quench[PwmCnt];
	   nxt   = PwmRead->Time[++PwmCnt];
       } while ((nxt <= TCNT0) && (nxt != 0xFF));
   PORTA &= buf; 
   }
OCR0= nxt;	
}